#!/usr/bin/env python
#
# Error types for pymwalib
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Adapted from:
# http://jakegoulding.com/rust-ffi-omnibus/objects/
#
# Additional documentation:
# https://docs.python.org/3.8/library/ctypes.html#module-ctypes
#

class ContextError(Exception):
    """Base class for other exceptions"""
    pass


class ContextContextNewError(ContextError):
    """Raised when call to C mwalibContext_get fails"""
    pass


class ContextAntennasGetError(ContextError):
    """Raised when call to C mwalib_antennas_get fails"""
    pass


class ContextCorrelatorBaselinesGetError(ContextError):
    """Raised when call to C mwalib_correlator_baselines_get fails"""
    pass


class ContextCorrelatorCoarseChannelsGetError(ContextError):
    """Raised when call to C mwalib_correlator_coarse_channels_get fails"""
    pass

class ContextVoltageCoarseChannelsGetError(ContextError):
    """Raised when call to C mwalib_voltage_coarse_channels_get fails"""
    pass

class ContextCoarseChannelsGetError(ContextError):
    """Raised when no context is provided"""
    pass

class ContextMetafitsMetadataGetError(ContextError):
    """Raised when call to C mwalib_metafits_metadata_get fails"""
    pass

class ContextCorrelatorMetadataGetError(ContextError):
    """Raised when call to C mwalib_correlator_metadata_get fails"""
    pass

class ContextVoltageMetadataGetError(ContextError):
    """Raised when call to C mwalib_voltage_metadata_get fails"""
    pass

class ContextRFInputsGetError(ContextError):
    """Raised when call to C mwalib_rfinputs_get fails"""
    pass

class ContextCorrelatorTimeStepsGetError(ContextError):
    """Raised when call to C mwalib_correlator_timesteps_get fails"""
    pass

class ContextVoltageTimeStepsGetError(ContextError):
    """Raised when call to C mwalib_voltage_timesteps_get fails"""
    pass

class ContextTimeStepsGetError(ContextError):
    """Raised when no context is provided"""
    pass

class ContextCorrelatorVisibilityPolsGetError(ContextError):
    """Raised when call to C mwalib_correlator_visibility_pols_get fails"""
    pass

class ContextMetafitsContextDisplayError(ContextError):
    """Raised when call to C mwalib_metafits_context_display fails"""
    pass

class ContextCorrelatorContextDisplayError(ContextError):
    """Raised when call to C mwalib_correlator_context_display fails"""
    pass

class ContextVoltageContextDisplayError(ContextError):
    """Raised when call to C mwalib_voltage_context_display fails"""
    pass

class ContextCorrelatorContextReadByBaselineException(ContextError):
    """Raised when call to C mwalibContext_read_by_baseline fails"""
    pass

class ContextCorrelatorContextReadByFrequencyException(ContextError):
    """Raised when call to C mwalibContext_read_by_frequency fails"""
    pass


#!/usr/bin/env python
#
# context: main interface for pymwalib for voltage data
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Adapted from:
# http://jakegoulding.com/rust-ffi-omnibus/objects/
#
# Additional documentation:
# https://docs.python.org/3.8/library/ctypes.html#module-ctypes
#
from pymwalib.mwalib import *
from pymwalib.common import ERROR_MESSAGE_LEN
from pymwalib.errors import *
from pymwalib.antenna import Antenna
from pymwalib.coarse_channel import CoarseChannel
from pymwalib.metafits_metadata import MetafitsMetadata
from pymwalib.voltage_metadata import VoltageMetadata
from pymwalib.rfinput import RFInput
from pymwalib.timestep import TimeStep


class VoltageContext:
    """Main class to interface with mwalib"""
    def __init__(self, metafits_filename: str, voltage_filenames: list):
        """Take metafits and voltage files, and populate this class via mwalib"""
        self._voltage_context_object = ct.POINTER(CVoltageContextS)()

        # First populate the context object
        self._get_voltage_context(metafits_filename, voltage_filenames)

        # Now Get metafits metadata
        self.metafits_metadata = MetafitsMetadata(None, None, self._voltage_context_object)

        # Now Get correlator metadata
        self.voltage_metadata = VoltageMetadata(self._voltage_context_object)

        # Populate rf_inputs
        self.rfinputs = RFInput.get_rfinputs(None, None, self._voltage_context_object)

        # Populate antennas
        self.ants = Antenna.get_antennas(None, None, self._voltage_context_object, self.rfinputs)

        # Populate coarse channels
        self.coarse_channels = CoarseChannel.get_coarse_channels(None, self._voltage_context_object)

        # Populate timesteps
        self.timesteps = TimeStep.get_timesteps(None, self._voltage_context_object)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Ensure we free the rust allocated memory in mwalib"""
        if self._voltage_context_object:
            mwalib.mwalib_voltage_context_free(self._voltage_context_object)

    def _get_voltage_context(self, metafits_filename: str, voltage_filenames: list):
        """This method will read and validate the metafits and voltage files. If all has worked, then
        the context object can be used in subsequent calls to populate aspects of this class."""
        if mwalib:
            # Encode all inputs as UTF-8.
            m = ct.c_char_p(metafits_filename.encode("utf-8"))

            # https://stackoverflow.com/questions/4145775/how-do-i-convert-a-python-list-into-a-c-array-by-using-ctypes
            encoded = []
            for v in voltage_filenames:
                encoded.append(ct.c_char_p(v.encode("utf-8")))
            seq = ct.c_char_p * len(encoded)
            g = seq(*encoded)
            error_message: bytes = create_string_buffer(ERROR_MESSAGE_LEN)

            if mwalib.mwalib_voltage_context_new(
                m, g, len(encoded), ct.byref(self._voltage_context_object), error_message, ERROR_MESSAGE_LEN) != 0:
                raise ContextContextNewError(f"Error creating voltage context object: "
                                             f"{error_message.decode('utf-8').rstrip()}")
        else:
            raise ContextContextNewError(f"Error creating voltage context object: libmwalib.so is not loaded.")

    def display(self):
         """Displays a human readable summary of the voltage context"""
         error_message = create_string_buffer(ERROR_MESSAGE_LEN)

         if mwalib.mwalib_voltage_context_display(self._voltage_context_object,
                                                  error_message,
                                                  ERROR_MESSAGE_LEN) != 0:
             raise ContextCorrelatorContextDisplayError(f"Error calling mwalib_voltage_context_display(): "
                                                        f"{error_message.decode('utf-8').rstrip()}")




#!/usr/bin/env python
#
# Common classes and functions
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
import enum

ERROR_MESSAGE_LEN = 1024

class CorrelatorVersion(enum.Enum):
    """Enum for Correlator version"""
    # MWAX correlator(v2)
    V2 = 0
    # MWA Legacy correlator(v1), having data files with "gpubox" and batch numbers in their names.
    Legacy = 1
    # MWA Old Legacy correlator(v1), having data files without any batch numbers.
    OldLegacy = 2


#!/usr/bin/env python
#
# visibility_pol: class representing a single visibility_pol in mwalib
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Adapted from:
# http://jakegoulding.com/rust-ffi-omnibus/objects/
#
# Additional documentation:
# https://docs.python.org/3.8/library/ctypes.html#module-ctypes
#
from pymwalib.mwalib import create_string_buffer, mwalib, CVisibilityPolS, CMetafitsContextS, CCorrelatorContextS, CVoltageContextS
from pymwalib.common import ERROR_MESSAGE_LEN
from pymwalib.errors import *
import ctypes as ct


class VisibilityPol:
    """
    A class representing a single mwalibVisibilityPol

    Attributes
    ----------
    index : int
        Ordinal index of this visibility pol.

    polarisation : str
        The polarisation of this object- e.g. "XX", "XY", "YX" or "YY"

    """

    def __init__(self,
                 index: int,
                 polarisation: str):
        """Initialise the class"""
        self.index: int = index
        self.polarisation: str = polarisation

    def __repr__(self):
        """Returns a representation of the class"""
        return f"{self.__class__.__name__}(" \
               f"Order: {self.index}, " \
               f"UNIX time: {self.polarisation})"

    @staticmethod
    def get_visibility_pols(metafits_context: ct.POINTER(CMetafitsContextS),
                            correlator_context: ct.POINTER(CCorrelatorContextS),
                            voltage_context: ct.POINTER(CVoltageContextS)) -> []:
        """Retrieve all of the visibility_pol metadata and populate a list of visibility_pols."""
        visibility_pols = []
        error_message: bytes = create_string_buffer(ERROR_MESSAGE_LEN)

        c_array_ptr = ct.POINTER(CVisibilityPolS)()
        c_len_ptr = ct.c_size_t(0)

        if mwalib.mwalib_visibility_pols_get(metafits_context,
                                             correlator_context,
                                             voltage_context,
                                             ct.byref(c_array_ptr),
                                             ct.byref(c_len_ptr),
                                             error_message,
                                             ERROR_MESSAGE_LEN) != 0:
            # Error
            raise ContextCorrelatorVisibilityPolsGetError(f"Error getting visibility_pol object: "
                                                          f"{error_message.decode('utf-8').rstrip()}")
        else:
            for i in range(0, c_len_ptr.value):
                # Populate all the fields
                visibility_pols.append(VisibilityPol(i, c_array_ptr[i].polarisation.decode("utf-8")))

            # We're now finished with the C memory, so free it
            mwalib.mwalib_visibility_pols_free(c_array_ptr, c_len_ptr.value)

            return visibility_pols


#!/usr/bin/env python
#
# context: main interface for pymwalib for just viewing metafits data (no data files)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Adapted from:
# http://jakegoulding.com/rust-ffi-omnibus/objects/
#
# Additional documentation:
# https://docs.python.org/3.8/library/ctypes.html#module-ctypes
#
from pymwalib.mwalib import *
from pymwalib.common import ERROR_MESSAGE_LEN
from pymwalib.errors import *
from pymwalib.antenna import Antenna
from pymwalib.baseline import Baseline
from pymwalib.metafits_metadata import MetafitsMetadata
from pymwalib.rfinput import RFInput
from pymwalib.visibility_pol import VisibilityPol


class MetafitsContext:
    """Main class to interface with mwalib"""
    def __init__(self, metafits_filename: str):
        """Take metafits and populate this class via mwalib"""
        self._metafits_context_object = ct.POINTER(CMetafitsContextS)()

        # First populate the context object
        self._get_metafits_context(metafits_filename)

        # Now Get metafits metadata
        self.metafits_metadata = MetafitsMetadata(self._metafits_context_object,None, None)

        # Populate rf_inputs
        self.rfinputs = RFInput.get_rfinputs(self._metafits_context_object, None, None)

        # Populate antennas
        self.ants = Antenna.get_antennas(self._metafits_context_object, None, None, self.rfinputs)

        # Populate baselines
        self.baselines = Baseline.get_baselines(self._metafits_context_object, None, None)

        # Populate visibility pols
        self.vis_pols = VisibilityPol.get_visibility_pols(self._metafits_context_object, None, None)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Ensure we free the rust allocated memory in mwalib"""
        if self._metafits_context_object:
            mwalib.mwalib_metafits_context_free(self._metafits_context_object)

    def _get_metafits_context(self, metafits_filename: str):
        """This method will read and validate the metafits and gpubox files. If all has worked, then
        the context object can be used in subsequent calls to populate aspects of this class."""
        if mwalib:
            # Encode all inputs as UTF-8.
            m = ct.c_char_p(metafits_filename.encode("utf-8"))

            error_message: bytes = create_string_buffer(ERROR_MESSAGE_LEN)

            if mwalib.mwalib_metafits_context_new(
                m, ct.byref(self._metafits_context_object), error_message, ERROR_MESSAGE_LEN) != 0:
                raise ContextContextNewError(f"Error creating metafits context object: "
                                             f"{error_message.decode('utf-8').rstrip()}")
        else:
            raise ContextContextNewError(f"Error creating correlator context object: libmwalib.so is not loaded.")

    def display(self):
         """Displays a human readable summary of the metafits context"""
         error_message = create_string_buffer(ERROR_MESSAGE_LEN)

         if mwalib.mwalib_metafits_context_display(self._metafits_context_object,
                                                   error_message,
                                                   ERROR_MESSAGE_LEN) != 0:
             raise ContextMetafitsContextDisplayError(f"Error calling mwalib_metafits_context_display(): "
                                                        f"{error_message.decode('utf-8').rstrip()}")



#!/usr/bin/env python
#
# rf_input: class representing a single rf_input in mwalib
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Adapted from:
# http://jakegoulding.com/rust-ffi-omnibus/objects/
#
# Additional documentation:
# https://docs.python.org/3.8/library/ctypes.html#module-ctypes
#
from pymwalib.mwalib import create_string_buffer, mwalib, CRFInputS, CMetafitsContextS, CCorrelatorContextS, CVoltageContextS
from pymwalib.common import ERROR_MESSAGE_LEN
from pymwalib.errors import *
import ctypes as ct


class RFInput:
    """
    A class representing a single RFInput

    Attributes
    ----------
    index : int
        Ordinal index of this RF Input.

    Please see https://github.com/MWATelescope/mwalib/blob/master/src/rf_input.rs for remaining attributes

    """

    def __init__(self,
                 index: int,
                 input: int,
                 ant: int,
                 tile_id: int,
                 tile_name: str,
                 pol: str,
                 electrical_length_m: float,
                 north_m: float,
                 east_m: float,
                 height_m: float,
                 vcs_order: int,
                 subfile_order: int,
                 flagged: bool,
                 rec_number: int,
                 rec_slot_number: int):
        """Initialise the class"""
        self.index: int = index
        self.input: int = input
        self.ant: int = ant
        self.tile_id: int = tile_id
        self.tile_name: str = tile_name
        self.pol: str = pol
        self.electrical_length_m: float = electrical_length_m
        self.north_m: float = north_m
        self.east_m: float = east_m
        self.height_m: float = height_m
        self.vcs_order: int = vcs_order
        self.subfile_order: int = subfile_order
        self.flagged: bool = flagged
        self.rec_number = rec_number
        self.rec_slot_number = rec_slot_number

    def __repr__(self):
        """Returns a representation of the class"""
        return f"{self.__class__.__name__}(" \
               f"Index: {self.index}, " \
               f"Input: {self.input}, " \
               f"Ant: {self.ant}, " \
               f"Tile Id: {self.tile_id}, " \
               f"Tile Name: {self.tile_name}, " \
               f"Pol: {self.pol}, " \
               f"Electrical length: {self.electrical_length_m}, " \
               f"north_m: {self.north_m}, " \
               f"east_m: {self.east_m}, " \
               f"height_m: {self.height_m}, " \
               f"vcs_order: {self.vcs_order}, " \
               f"subfile_order: {self.subfile_order}, " \
               f"flagged: {self.flagged}, " \
               f"rec_number: {self.rec_number}, " \
               f"rec_slot_number: {self.rec_slot_number})"

    @staticmethod
    def get_rfinputs(metafits_context: ct.POINTER(CMetafitsContextS),
                     correlator_context: ct.POINTER(CCorrelatorContextS),
                     voltage_context: ct.POINTER(CVoltageContextS)) -> []:
        """Retrieve all of the rf_input metadata and populate a list of rf_inputs."""
        rf_inputs = []
        error_message: bytes = create_string_buffer(ERROR_MESSAGE_LEN)

        c_array_ptr = ct.POINTER(CRFInputS)()
        c_len_ptr = ct.c_size_t(0)

        if mwalib.mwalib_rfinputs_get(metafits_context,
                                      correlator_context,
                                      voltage_context,
                                      ct.byref(c_array_ptr),
                                      ct.byref(c_len_ptr),
                                      error_message,
                                      ERROR_MESSAGE_LEN) != 0:
            # Error
            raise ContextRFInputsGetError(f"Error getting rf_inputs object: "
                                          f"{error_message.decode('utf-8').rstrip()}")
        else:
            for i in range(0, c_len_ptr.value):
                # Populate all the fields
                rf_inputs.append(RFInput(i,
                                         c_array_ptr[i].input,
                                         c_array_ptr[i].ant,
                                         c_array_ptr[i].tile_id,
                                         c_array_ptr[i].tile_name.decode("utf-8"),
                                         c_array_ptr[i].pol.decode("utf-8"),
                                         c_array_ptr[i].electrical_length_m,
                                         c_array_ptr[i].north_m,
                                         c_array_ptr[i].east_m,
                                         c_array_ptr[i].height_m,
                                         c_array_ptr[i].vcs_order,
                                         c_array_ptr[i].subfile_order,
                                         c_array_ptr[i].flagged,
                                         c_array_ptr[i].rec_number,
                                         c_array_ptr[i].rec_slot_number))

            # We're now finished with the C memory, so free it
            mwalib.mwalib_rfinputs_free(c_array_ptr, c_len_ptr.value)

            return rf_inputs


#!/usr/bin/env python
#
# Class for accessing Correlator Metadata
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Adapted from:
# http://jakegoulding.com/rust-ffi-omnibus/objects/
#
# Additional documentation:
# https://docs.python.org/3.8/library/ctypes.html#module-ctypes
#
from pymwalib.mwalib import *
from pymwalib.common import ERROR_MESSAGE_LEN,CorrelatorVersion
from pymwalib.errors import *


class CorrelatorMetadata:
    def __init__(self,
                 correlator_context: ct.POINTER(CCorrelatorContextS)
                 ):
        """Retrieve all of the correlator metadata and populate this class."""
        error_message: bytes = create_string_buffer(ERROR_MESSAGE_LEN)

        c_object_ptr = ct.POINTER(CCorrelatorMetadataS)()

        if mwalib.mwalib_correlator_metadata_get(correlator_context,
                                                 ct.byref(c_object_ptr),
                                                 error_message,
                                                 ERROR_MESSAGE_LEN) != 0:
            raise ContextCorrelatorMetadataGetError(
                f"Error creating correlator metadata object: {error_message.decode('utf-8').rstrip()}")
        else:
            c_object =c_object_ptr.contents

            # Populate all the fields
            self.corr_version: CorrelatorVersion = c_object.corr_version
            self.start_unix_time_ms: int = c_object.start_unix_time_ms
            self.end_unix_time_ms: int = c_object.end_unix_time_ms
            self.start_gps_time_ms: int = c_object.start_gps_time_ms
            self.end_gps_time_ms: int = c_object.end_gps_time_ms
            self.duration_ms: int = c_object.duration_ms
            self.num_timesteps: int = c_object.num_timesteps
            self.num_coarse_chans: int = c_object.num_coarse_chans
            self.bandwidth_hz: int = c_object.bandwidth_hz
            self.num_timestep_coarse_chan_bytes: int = c_object.num_timestep_coarse_chan_bytes
            self.num_timestep_coarse_chan_floats: int = c_object.num_timestep_coarse_chan_floats
            self.num_gpubox_files: int = c_object.num_gpubox_files

            # We're now finished with the C memory, so free it
            mwalib.mwalib_correlator_metadata_free(c_object)

    def __repr__(self):
        """Returns a representation of the class"""
        return f"{self.__class__.__name__}(\n" \
               f"Correlator Version                    : {CorrelatorVersion(self.corr_version).name}\n" \
               f"(actual) Start time (UNIX)            : {float(self.start_unix_time_ms) / 1000.} UNIX\n" \
               f"(actual) End time (UNIX)              : {float(self.end_unix_time_ms) / 1000.} UNIX\n" \
               f"(actual) Start time (GPS)             : {float(self.start_gps_time_ms) / 1000.} UNIX\n" \
               f"(actual) End time (GPS)               : {float(self.end_gps_time_ms) / 1000.} UNIX\n" \
               f"(actual) Duration                     : {float(self.duration_ms) / 1000.} s\n" \
               f"(actual) num timesteps                : {self.num_timesteps}\n" \
               f"(actual) num coarse channels          : {self.num_coarse_chans}\n" \
               f"(Data) Bandwidth (of data we have)    : {float(self.bandwidth_hz) / 1000000.} MHz\n" \
               f"Num bytes per timestep coarse channel : " \
               f"{float(self.num_timestep_coarse_chan_bytes) / (1024. * 1024.)} MB\n" \
               f"Num floats per timestep coarse channel: " \
               f"{self.num_timestep_coarse_chan_floats}\n" \
               f"(actual) num GPUBox files             : {self.num_gpubox_files}\n)\n"


#!/usr/bin/env python
#
# timestep: class representing a single timestep in mwalib
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Adapted from:
# http://jakegoulding.com/rust-ffi-omnibus/objects/
#
# Additional documentation:
# https://docs.python.org/3.8/library/ctypes.html#module-ctypes
#
from pymwalib.mwalib import create_string_buffer, mwalib, CTimeStepS, CCorrelatorContextS, CVoltageContextS
from pymwalib.common import ERROR_MESSAGE_LEN
from pymwalib.errors import *
import ctypes as ct


class TimeStep:
    """
    A class representing a single TimeStep

    Attributes
    ----------
    index : int
        Ordinal index of this time step.

    unix_time_ms : int
        The UNIX time (in milliseconds) of the start of this time step

    gps_time_ms : int
        The GPS time (in milliseconds) of the start of this time step

    """

    def __init__(self,
                 index: int,
                 unix_time_ms: int,
                 gps_time_ms: int):
        """Initialise the class"""
        self.index: int = index
        self.unix_time_ms: int = unix_time_ms
        self.gps_time_ms: int = gps_time_ms

    def __repr__(self):
        """Returns a representation of the class"""
        return f"{self.__class__.__name__}(" \
               f"Order: {self.index}, " \
               f"UNIX time: {float(self.unix_time_ms) / 1000.}, " \
               f"GPS time: {float(self.gps_time_ms) / 1000.})"

    @staticmethod
    def get_timesteps(correlator_context: ct.POINTER(CCorrelatorContextS),
                      voltage_context: ct.POINTER(CVoltageContextS)) -> []:
        """Retrieve all of the timestep metadata and populate a list of timesteps."""
        timesteps = []
        error_message: bytes = create_string_buffer(ERROR_MESSAGE_LEN)

        c_array_ptr = ct.POINTER(CTimeStepS)()
        c_len_ptr = ct.c_size_t(0)

        if correlator_context is not None:
            if mwalib.mwalib_correlator_timesteps_get(correlator_context,
                                                      ct.byref(c_array_ptr),
                                                      ct.byref(c_len_ptr),
                                                      error_message,
                                                      ERROR_MESSAGE_LEN) != 0:
                # Error
                raise ContextCorrelatorTimeStepsGetError(f"Error getting timestep object: "
                                                         f"{error_message.decode('utf-8').rstrip()}")
        elif voltage_context is not None:
            if mwalib.mwalib_voltage_timesteps_get(voltage_context,
                                                   ct.byref(c_array_ptr),
                                                   ct.byref(c_len_ptr),
                                                   error_message,
                                                   ERROR_MESSAGE_LEN) != 0:
                # Error
                raise ContextVoltageTimeStepsGetError(f"Error getting timestep object: "
                                                      f"{error_message.decode('utf-8').rstrip()}")
        else:
            raise ContextTimeStepsGetError(f"Error getting timestep object: "
                                           f"neither correlator nor voltage context provided.")

        for i in range(0, c_len_ptr.value):
            # Populate all the fields
            timesteps.append(TimeStep(i,
                                      c_array_ptr[i].unix_time_ms,
                                      c_array_ptr[i].gps_time_ms))

        # We're now finished with the C memory, so free it
        mwalib.mwalib_timesteps_free(c_array_ptr, c_len_ptr.value)

        return timesteps


#!/usr/bin/env python
#
# context: main interface for pymwalib for correlator data
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Adapted from:
# http://jakegoulding.com/rust-ffi-omnibus/objects/
#
# Additional documentation:
# https://docs.python.org/3.8/library/ctypes.html#module-ctypes
#
import numpy.ctypeslib as npct
from pymwalib.mwalib import *
from pymwalib.common import ERROR_MESSAGE_LEN
from pymwalib.errors import *
from pymwalib.antenna import Antenna
from pymwalib.baseline import Baseline
from pymwalib.coarse_channel import CoarseChannel
from pymwalib.metafits_metadata import MetafitsMetadata
from pymwalib.correlator_metadata import CorrelatorMetadata
from pymwalib.rfinput import RFInput
from pymwalib.timestep import TimeStep
from pymwalib.visibility_pol import VisibilityPol


class CorrelatorContext:
    """Main class to interface with mwalib"""
    def __init__(self, metafits_filename: str, gpubox_filenames: list):
        """Take metafits and gpubox files, and populate this class via mwalib"""
        self._correlator_context_object = ct.POINTER(CCorrelatorContextS)()

        # First populate the context object
        self._get_correlator_context(metafits_filename, gpubox_filenames)

        # Now Get metafits metadata
        self.metafits_metadata = MetafitsMetadata(None, self._correlator_context_object, None)

        # Now Get correlator metadata
        self.correlator_metadata = CorrelatorMetadata(self._correlator_context_object)

        # Populate rf_inputs
        self.rfinputs = RFInput.get_rfinputs(None, self._correlator_context_object, None)

        # Populate antennas
        self.antennas = Antenna.get_antennas(None, self._correlator_context_object, None, self.rfinputs)

        # Populate baselines
        self.baselines = Baseline.get_baselines(None, self._correlator_context_object, None)

        # Populate coarse channels
        self.coarse_channels = CoarseChannel.get_coarse_channels(self._correlator_context_object, None)

        # Populate timesteps
        self.timesteps = TimeStep.get_timesteps(self._correlator_context_object, None)

        # Populate visibility pols
        self.visibility_pols = VisibilityPol.get_visibility_pols(None, self._correlator_context_object, None)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Ensure we free the rust allocated memory in mwalib"""
        if self._correlator_context_object:
            mwalib.mwalib_correlator_context_free(self._correlator_context_object)

    def _get_correlator_context(self, metafits_filename: str, gpubox_filenames: list):
        """This method will read and validate the metafits and gpubox files. If all has worked, then
        the context object can be used in subsequent calls to populate aspects of this class."""
        if mwalib:
            # Encode all inputs as UTF-8.
            m = ct.c_char_p(metafits_filename.encode("utf-8"))

            # https://stackoverflow.com/questions/4145775/how-do-i-convert-a-python-list-into-a-c-array-by-using-ctypes
            encoded = []
            for g in gpubox_filenames:
                encoded.append(ct.c_char_p(g.encode("utf-8")))
            seq = ct.c_char_p * len(encoded)
            g = seq(*encoded)
            error_message: bytes = create_string_buffer(ERROR_MESSAGE_LEN)

            if mwalib.mwalib_correlator_context_new(
                m, g, len(encoded), ct.byref(self._correlator_context_object), error_message, ERROR_MESSAGE_LEN) != 0:
                raise ContextContextNewError(f"Error creating correlator context object: "
                                             f"{error_message.decode('utf-8').rstrip()}")
        else:
            raise ContextContextNewError(f"Error creating correlator context object: libmwalib.so is not loaded.")

    def display(self):
         """Displays a human readable summary of the correlator context"""
         error_message = create_string_buffer(ERROR_MESSAGE_LEN)

         if mwalib.mwalib_correlator_context_display(self._correlator_context_object,
                                                     error_message,
                                                     ERROR_MESSAGE_LEN) != 0:
             raise ContextCorrelatorContextDisplayError(f"Error calling mwalib_correlator_context_display(): "
                                                        f"{error_message.decode('utf-8').rstrip()}")

    def read_by_baseline(self, timestep_index, coarse_chan_index):
         """Retrieve one HDU (ordered baseline,freq,pol,r,i) as a numpy array."""
         error_message = " ".encode("utf-8") * ERROR_MESSAGE_LEN

         float_buffer_type = ct.c_float * self.correlator_metadata.num_timestep_coarse_chan_floats
         buffer = float_buffer_type()

         if mwalib.mwalib_correlator_context_read_by_baseline(self._correlator_context_object,
                                                              ct.c_size_t(timestep_index),
                                                              ct.c_size_t(coarse_chan_index),
                                                              buffer,
                                                              self.correlator_metadata.num_timestep_coarse_chan_floats,
                                                              error_message, ERROR_MESSAGE_LEN) != 0:
             raise ContextCorrelatorContextReadByBaselineException(f"Error reading data: "
                                                  f"{error_message.decode('utf-8').rstrip()}")
         else:
             return npct.as_array(buffer, shape=(self.correlator_metadata.num_timestep_coarse_chan_floats,))

    def read_by_frequency(self, timestep_index, coarse_chan_index):
        """Retrieve one HDU (ordered freq,baseline,pol,r,i) as a numpy array."""
        error_message = " ".encode("utf-8") * ERROR_MESSAGE_LEN

        float_buffer_type = ct.c_float * self.correlator_metadata.num_timestep_coarse_chan_floats
        buffer = float_buffer_type()

        if mwalib.mwalib_correlator_context_read_by_frequency(self._correlator_context_object,
                                                              ct.c_size_t(timestep_index),
                                                              ct.c_size_t(coarse_chan_index),
                                                              buffer,
                                                              self.correlator_metadata.num_timestep_coarse_chan_floats,
                                                              error_message, ERROR_MESSAGE_LEN) != 0:
            raise ContextCorrelatorContextReadByFrequencyException(f"Error reading data: "
                                                                   f"{error_message.decode('utf-8').rstrip()}")
        else:
            return npct.as_array(buffer, shape=(self.correlator_metadata.num_timestep_coarse_chan_floats,))



#!/usr/bin/env python
#
# coarse_channel: class representing a single coarse_channel in mwalib
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Adapted from:
# http://jakegoulding.com/rust-ffi-omnibus/objects/
#
# Additional documentation:
# https://docs.python.org/3.8/library/ctypes.html#module-ctypes
#
from pymwalib.mwalib import create_string_buffer, mwalib, CCoarseChannelS, CCorrelatorContextS, CVoltageContextS
from pymwalib.common import ERROR_MESSAGE_LEN
from pymwalib.errors import *
import ctypes as ct


class CoarseChannel:
    """
    A class representing a single CoarseChannel

    Attributes
    ----------
    index : int
        Ordinal index of this coarse channel.

    Please see https://github.com/MWATelescope/mwalib/blob/master/src/coarse_channel.rs for remaining attributes

    """

    def __init__(self,
                 index: int,
                 corr_chan_number: int,
                 rec_chan_number: int,
                 gpubox_number: int,
                 chan_width_hz: int,
                 chan_start_hz: int,
                 chan_centre_hz: int,
                 chan_end_hz: int):
        """Initialise the class"""
        self.index: int = index
        self.corr_chan_number: int = corr_chan_number
        self.rec_chan_number: int = rec_chan_number
        self.gpubox_number: int = gpubox_number
        self.chan_width_hz: int = chan_width_hz
        self.chan_start_hz: int = chan_start_hz
        self.chan_centre_hz: int = chan_centre_hz
        self.chan_end_hz: int = chan_end_hz

    def __repr__(self):
        """Returns a representation of the class"""
        return f"{self.__class__.__name__}(" \
               f"Order: {self.index}, " \
               f"Correlator channel number: {self.corr_chan_number}, " \
               f"Receiver channel number: {self.rec_chan_number}, " \
               f"GPUBox Number: {self.gpubox_number}, " \
               f"Channel width MHz: {float(self.chan_width_hz) / 1000000.}, " \
               f"Channel start MHz: {float(self.chan_start_hz) / 1000000.}, " \
               f"Channel centre MHz: {float(self.chan_centre_hz) / 1000000.}, " \
               f"Channel end MHz: {float(self.chan_end_hz) / 1000000.})"

    @staticmethod
    def get_coarse_channels(correlator_context: ct.POINTER(CCorrelatorContextS),
                            voltage_context: ct.POINTER(CVoltageContextS)) -> []:
        """Retrieve all of the coarse_channel metadata and populate a list of coarse_channels."""
        coarse_channels = []
        error_message: bytes = create_string_buffer(ERROR_MESSAGE_LEN)

        c_array_ptr = ct.POINTER(CCoarseChannelS)()
        c_len_ptr = ct.c_size_t(0)

        if correlator_context is not None:
            if mwalib.mwalib_correlator_coarse_channels_get(correlator_context,
                                                      ct.byref(c_array_ptr),
                                                      ct.byref(c_len_ptr),
                                                      error_message,
                                                      ERROR_MESSAGE_LEN) != 0:
                # Error
                raise ContextCorrelatorCoarseChannelsGetError(f"Error getting coarse_channel object: "
                                                              f"{error_message.decode('utf-8').rstrip()}")
        elif voltage_context is not None:
             if mwalib.mwalib_voltage_coarse_channels_get(correlator_context,
                                                       ct.byref(c_array_ptr),
                                                       ct.byref(c_len_ptr),
                                                       error_message,
                                                       ERROR_MESSAGE_LEN) != 0:
                # Error
                raise ContextVoltageCoarseChannelsGetError(f"Error getting coarse_channel object: "
                                                           f"{error_message.decode('utf-8').rstrip()}")
        else:
            raise ContextCoarseChannelsGetError(f"Error getting coarse_channel object: "
                                                f"neither correlator nor voltage context provided.")

        for i in range(0, c_len_ptr.value):
            # Populate all the fields
            coarse_channels.append(CoarseChannel(i,
                                                 c_array_ptr[i].corr_chan_number,
                                                 c_array_ptr[i].rec_chan_number,
                                                 c_array_ptr[i].gpubox_number,
                                                 c_array_ptr[i].chan_width_hz,
                                                 c_array_ptr[i].chan_start_hz,
                                                 c_array_ptr[i].chan_centre_hz,
                                                 c_array_ptr[i].chan_end_hz,))

        # We're now finished with the C memory, so free it
        mwalib.mwalib_coarse_channels_free(c_array_ptr, c_len_ptr.value)

        return coarse_channels


#!/usr/bin/env python
#
# antenna: class representing a single antenna in mwalib
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Adapted from:
# http://jakegoulding.com/rust-ffi-omnibus/objects/
#
# Additional documentation:
# https://docs.python.org/3.8/library/ctypes.html#module-ctypes
#
from pymwalib.mwalib import create_string_buffer, mwalib, CAntennaS, CMetafitsContextS, CCorrelatorContextS, CVoltageContextS
from pymwalib.common import ERROR_MESSAGE_LEN
from pymwalib.errors import *
from pymwalib.rfinput import RFInput
import ctypes as ct

class Antenna:
    """
    A class representing a single mwalibAntenna

    Attributes
    ----------
    index : int
        Ordinal index of this RF Input.

    Please see https://github.com/MWATelescope/mwalib/blob/master/src/antenna.rs for remaining attributes

    """

    def __init__(self,
                 index: int,
                 ant: int,
                 tile_id: int,
                 tile_name: str,
                 rf_input_x: RFInput,
                 rf_input_y: RFInput):
        """Initialise the class"""
        self.index: int = index
        self.ant: int = ant
        self.tile_id: int = tile_id
        self.tile_name: str = tile_name
        self.rf_input_x: RFInput = rf_input_x
        self.rf_input_y: RFInput = rf_input_y

    def __repr__(self):
        """Returns a representation of the class"""
        return f"{self.__class__.__name__}(" \
               f"Index: {self.index}, " \
               f"ant: {self.ant}, " \
               f"Tile Id: {self.tile_id}, " \
               f"Tile Name: {self.tile_name}, " \
               f"RF Input X: {self.rf_input_x!r}, " \
               f"RF Input Y: {self.rf_input_y!r})"

    @staticmethod
    def get_antennas(metafits_context: ct.POINTER(CMetafitsContextS),
                     correlator_context: ct.POINTER(CCorrelatorContextS),
                     voltage_context: ct.POINTER(CVoltageContextS),
                     rf_inputs: []) -> []:
        """Retrieve all of the antenna metadata and populate a list of antennas."""
        antennas = []
        error_message: bytes = create_string_buffer(ERROR_MESSAGE_LEN)

        c_array_ptr = ct.POINTER(CAntennaS)()
        c_len_ptr = ct.c_size_t(0)

        if mwalib.mwalib_antennas_get(metafits_context,
                                      correlator_context,
                                      voltage_context,
                                      ct.byref(c_array_ptr),
                                      ct.byref(c_len_ptr),
                                      error_message,
                                      ERROR_MESSAGE_LEN) != 0:
            # Error
            raise ContextAntennasGetError(f"Error getting antennas object: "
                                          f"{error_message.decode('utf-8').rstrip()}")
        else:
            for i in range(0, c_len_ptr.value):
                # Populate all the fields
                antennas.append(Antenna(i,
                                 c_array_ptr[i].ant,
                                 c_array_ptr[i].tile_id,
                                 c_array_ptr[i].tile_name.decode("utf-8"),
                                 rf_inputs[c_array_ptr[i].rfinput_x],
                                 rf_inputs[c_array_ptr[i].rfinput_y]))

            # We're now finished with the C memory, so free it
            mwalib.mwalib_antennas_free(c_array_ptr, c_len_ptr.value)

            return antennas